use serde::{Serialize, Deserialize};
use std::fs::File;
use std::str::FromStr;
use std::io::prelude::*;
use std::collections::HashMap;
use log::{error, warn};

#[cfg(not(debug_assertions))]
macro_rules! PATH { () => { "/etc/mrproxy/" } }
#[cfg(not(debug_assertions))]
static FILE_CONF: &str = concat!(PATH!(),"mrprox.conf");
#[cfg(not(debug_assertions))]
static FILE_SERVERS: &str = concat!(PATH!(),"mrprox_servers.conf");

#[cfg(debug_assertions)]
static FILE_CONF: &str = "mrprox.conf";
#[cfg(debug_assertions)]
static FILE_SERVERS: &str = "mrprox_servers.conf";

const DEF_PORT: u16 = 25565;

const TCP_TYPE: &str = "TCP";
const UNIX_TYPE: &str = "UNIX";

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct ConfFile{
    ip: String,
    port: String,
    conf: ConfServer,
}

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct ConfServer{
    sock_type: String,
    path: Option<String>,
    ip: Option<String>,
    port: Option<String>,
}

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct ServerData{
    pub domain: String,
    pub ip: String,
}

pub struct Config{
    l_servers : HashMap<String, String>,
    port: String,
    ip: String,
    conf: ConfServer,
}

impl Config {
    pub fn new() -> Self {
        let mut conf_file = File::open(&FILE_CONF).unwrap();
        let mut servers_file = File::open(&FILE_SERVERS).unwrap();
        let mut s1 = String::new();
        let mut s2 = String::new();
        conf_file.read_to_string(&mut s1).unwrap();
        servers_file.read_to_string(&mut s2).unwrap();
        let yam_conf:ConfFile = match serde_yaml::from_str(&s1) {
            Ok(result) => result,
            Err(e) => {error!("Config file malformed: {}",e); panic!("{}", e);},
        };
        let yam_ser: Vec<ServerData> = match serde_yaml::from_str(&s2) {
            Ok(result) => result,
            Err(e) => {warn!("IPs list file malformed; creating new one"); Vec::new()},
        };
        Self{
            l_servers: Self::get_servers(&yam_ser),
            port: yam_conf.port,
            ip: yam_conf.ip,
            conf: yam_conf.conf,
        }
    }

    pub fn add(&mut self, server: ServerData) -> bool{
        match self.l_servers.insert(server.domain, server.ip){
            Some(_s) => true, //Modified instead of added
            None => false,
        }
    }

    pub fn del(&mut self, domain: String) -> bool {
        match self.l_servers.remove(&domain) {
            Some(_s) => true,
            None => false,
        }
    }

    pub fn flush(&self){
        let servers = Vec::from_iter(self.l_servers.iter()
                    .map(|(x, y)| ServerData{domain: (*x).clone(), ip: (*y).clone()}));
        let mut file = File::create(FILE_SERVERS).unwrap();
        file.write(&serde_yaml::to_string(&servers).unwrap().into_bytes());
        file.flush();

    }

    fn get_servers(file: &Vec<ServerData>) -> HashMap<String, String> {
        let mut ret = HashMap::new();
        for j in file{
            ret.insert(j.domain.clone(),j.ip.clone());
        }
        ret
    }

    pub fn get_host(&self, host: &String) -> Option<(&String, &String)>{
        self.l_servers.get_key_value(host)
    }

    pub fn get_server(&self, host: &String) -> Option<(String, u16)>{
        let raw: Vec<&str> =
        match self.get_host(host){
            Some(h) => h.1.split(":").collect(),
            None => return None,
        };

        match raw.len() {
            1 => Some((String::from(raw[0]),DEF_PORT)),
            2 => {
                match FromStr::from_str(raw[1]) {
                    Ok(p) => Some((String::from(raw[0]),p)),
                    Err(_e) => None,
                }
            },
            _=> None,
        }
    }

    pub fn get_port(&self) -> &String{
        &self.port
    }

    pub fn get_ip(&self) -> &String{
        &self.ip
    }

    pub fn get_port_conf(&self) -> &String{
        match self.conf.sock_type.as_str() {
            TCP_TYPE => &self.conf.port.as_ref().unwrap(),
            _=> {error!("Only tcp types have port"); panic!("Only tcp types have port")},
        }
    }

    pub fn get_conf_ip(&self) -> &String{
        match self.conf.sock_type.as_str() {
            TCP_TYPE => &self.conf.ip.as_ref().unwrap(),
            _=> {error!("Only tcp types have IP"); panic!("Only tcp types have IP")},
        }
    }

    pub fn get_bindeable_ip(&self) -> String {
        build_bindeable_ip(&self.ip,&self.port)
    }

    pub fn get_bindeable_conf(&self) -> String {
        match self.conf.sock_type.as_str() {
            TCP_TYPE => build_bindeable_ip(
                            self.conf.ip.as_ref().unwrap(),
                            self.conf.port.as_ref().unwrap()
                        ),
            UNIX_TYPE => self.conf.path.as_ref().unwrap().to_string(),
            _=> {error!("Invalid type"); panic!("Invalid type")},
        }
    }

    pub fn get_conf_type(&self) -> &String{
        &self.conf.sock_type
    }
}

fn build_bindeable_ip(ip: &String, port: &String) -> String{
    format!("{}:{}",ip,port)
}

